#
# zmey Project
# Copyright (C) 2018 by Contributors <https://github.com/Tyill/zmey>
#
# This code is licensed under the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, sbject to the following conditions :
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# coding: utf-8

from __future__ import absolute_import
import ctypes
from typing import List
from enum import Enum


_lib = None

def loadLib(path : str):
  global _lib
  _lib = ctypes.CDLL(path)
  
#############################################################################
### Common

class StateType(Enum):
  """State type"""
  UNDEFINED      = -1
  READY          = 0
  START          = 1
  RUNNING        = 2
  PAUSE          = 3
  STOP           = 4    
  COMPLETED      = 5
  ERROR          = 6
  CANCEL         = 7
  NOT_RESPONDING = 8

def stateStr(state : int):
  return StateType(state).name 

class Scheduler: 
  """Scheduler config""" 
  def __init__(self,
               id : int = 0, 
               state : int = 0,
               connectPnt : str = "",
               capacityTask : int = 10000,
               activeTask : int = 0,
               startTime : str = "",
               stopTime : str = "",
               pingTime : str = "",
               name = "",
               description = ""):
    self.id = id
    self.state = state
    self.connectPnt = connectPnt     # remote connection point: IP or DNS:port
    self.capacityTask = capacityTask # permissible simultaneous number of tasks 
    self.activeTask  = activeTask
    self.startTime  = startTime
    self.stopTime  = stopTime
    self.pingTime  = pingTime
    self.name = name
    self.description = description
    def __repr__(self):
      return f"Scheduler: id {self.id} state {self.state} connectPnt {self.connectPnt} \
                       capacityTask {self.capacityTask} activeTask {self.activeTask} pingTime {self.pingTime} \
                       startTime {self.startTime} stopTime {self.stopTime} name {self.name} description {self.description}"
    def __str__(self):
      return self.__repr__()
class Worker: 
  """Worker config""" 
  def __init__(self,
               id : int = 0,
               sId : int = 0, 
               state : int = 0,
               connectPnt : str = "",
               capacityTask : int = 10,
               activeTask : int = 0,
               load : int = 0,
               startTime : str = "",
               stopTime : str = "",
               pingTime : str = "",
               name = "",
               description = ""):
    self.id = id
    self.sId = sId                   # Scheduler id
    self.state = state
    self.connectPnt = connectPnt     # remote connection point: IP or DNS:port
    self.capacityTask = capacityTask # permissible simultaneous number of tasks 
    self.activeTask  = activeTask
    self.load  = load
    self.startTime  = startTime
    self.stopTime  = stopTime
    self.pingTime  = pingTime
    self.name = name
    self.description = description
    def __repr__(self):
      return f"Worker: id {self.id} sId {self.sId} state {self.state} connectPnt {self.connectPnt} \
                       capacityTask {self.capacityTask} activeTask {self.activeTask} load {self.load} \
                       startTime {self.startTime} stopTime {self.stopTime} name {self.name} description {self.description}"
    def __str__(self):
      return self.__repr__()
class TaskTemplate: 
  """TaskTemplate config""" 
  def __init__(self,
               id : int = 0,
               userId : int = 0,
               averDurationSec : int = 1,
               maxDurationSec : int = 1,
               name : str = "",
               description : str = "",
               script: str = ""):
    self.id = id
    self.userId = userId               
    self.averDurationSec = averDurationSec
    self.maxDurationSec = maxDurationSec
    self.name = name    
    self.description = description
    self.script = script   
  def __repr__(self):
      return f"TaskTemplate: name {self.name} description {self.description} script {self.script} id {self.id} userId {self.userId} averDurationSec {self.averDurationSec} maxDurationSec {self.maxDurationSec}   "
  def __str__(self):
      return self.__repr__()

class Task:
  """Task config""" 
  def __init__(self,
               id : int = 0,         
               ttlId : int = 0,         # Task template id    
               state : int = 0,
               progress : int = 0,
               result : str = "",
               params : str = "",
               createTime : str = "",
               takeInWorkTime : str = "",
               startTime : str = "",
               stopTime : str = ""):
    self.id = id
    self.ttlId = ttlId
    self.state = state
    self.progress = progress
    self.result = result
    self.params = params               # CLI params for script
    self.createTime = createTime
    self.takeInWorkTime = takeInWorkTime
    self.startTime = startTime
    self.stopTime = stopTime
    def __repr__(self):
      return f"Task: id {self.id} ttlId {self.ttlId} state {self.state} stopTime {self.stopTime} startTime {self.startTime} takeInWorkTime {self.takeInWorkTime} createTime {self.createTime} params {self.params} progress {self.progress} result {self.result}"
    def __str__(self):
      return self.__repr__()
class InternError: 
  """Internal ERROR""" 
  def __init__(self,
               sId : int = 0,
               wId : int = 0,
               createTime : str = 0,
               message : str = ""):
    self.schedrId = sId                 # Scheduler id    
    self.workerId = wId                 # Worker id
    self.createTime = createTime   
    self.message = message          
  
class _ConnectCng_C(ctypes.Structure):
  _fields_ = [('connStr', ctypes.c_char_p)]
class _SchedrCng_C(ctypes.Structure):
  _fields_ = [('capacityTask', ctypes.c_uint32),
              ('connectPnt', ctypes.c_char * 256),
              ('name', ctypes.c_char * 256),
              ('description', ctypes.c_char_p)]
class _SchedrState_C(ctypes.Structure):
  _fields_ = [('state', ctypes.c_uint32),
              ('activeTask', ctypes.c_uint32),
              ('startTime', ctypes.c_char * 32),
              ('stopTime', ctypes.c_char * 32),
              ('pingTime', ctypes.c_char * 32)]
class _WorkerCng_C(ctypes.Structure):
  _fields_ = [('schedrId', ctypes.c_uint64),
              ('capacityTask', ctypes.c_uint32),
              ('connectPnt', ctypes.c_char * 256),
              ('name', ctypes.c_char * 256),
              ('description', ctypes.c_char_p)]
class _WorkerState_C(ctypes.Structure):
  _fields_ = [('state', ctypes.c_uint32),
              ('activeTask', ctypes.c_uint32),
              ('load', ctypes.c_uint32),
              ('startTime', ctypes.c_char * 32),
              ('stopTime', ctypes.c_char * 32),
              ('pingTime', ctypes.c_char * 32)]
class _TaskTemplCng_C(ctypes.Structure):
  _fields_ = [('userId', ctypes.c_uint64),
              ('schedrPresetId', ctypes.c_uint64),
              ('workerPresetId', ctypes.c_uint64),
              ('averDurationSec', ctypes.c_uint32),
              ('maxDurationSec', ctypes.c_uint32),
              ('name', ctypes.c_char * 256),
              ('description', ctypes.c_char_p),
              ('script', ctypes.c_char_p)]
class _TaskCng_C(ctypes.Structure):
  _fields_ = [('ttlId', ctypes.c_uint64),
              ('params', ctypes.c_char_p)]
class _TaskState_C(ctypes.Structure):
  _fields_ = [('progress', ctypes.c_uint32),
              ('state', ctypes.c_int32)]
class _TaskTime_C(ctypes.Structure):
  _fields_ = [('createTime', ctypes.c_char * 32),
              ('takeInWorkTime', ctypes.c_char * 32),
              ('startTime', ctypes.c_char * 32),
              ('stopTime', ctypes.c_char * 32)]
class _InternError_C(ctypes.Structure):
  _fields_ = [('schedrId', ctypes.c_uint64),
              ('workerId', ctypes.c_uint64),
              ('createTime', ctypes.c_char * 32),
              ('message', ctypes.c_char * 256)]

def version() -> str:
  """
  Version library
  :return: version
  """
  pfun = _lib.zmVersionLib
  pfun.restype = None
  pfun.argtypes = (ctypes.c_char_p,)

  ver = ctypes.create_string_buffer(32)
  pfun(ver)
  return ver.value.decode("utf-8")

class Connection:
  """Connection object"""
  
  _zmConn = None
  _userErrCBack : ctypes.CFUNCTYPE(None, ctypes.c_char_p, ctypes.c_void_p) = None
  _changeTaskStateCBack : ctypes.CFUNCTYPE(None, ctypes.c_uint64, ctypes.c_uint64, ctypes.c_int32, ctypes.c_int32, ctypes.c_void_p) = None
  _cerr : str = ""
  
  def __init__(self, connStr : str):
    if not _lib:
      raise Exception('lib not load')
    
    cng = _ConnectCng_C()
    cng.connStr = connStr.encode("utf-8")
    
    pfun = _lib.zmCreateConnection
    pfun.argtypes = (_ConnectCng_C, ctypes.c_char_p)
    pfun.restype = ctypes.c_void_p
    err = ctypes.create_string_buffer(256)
    self._zmConn = pfun(cng, err)
    self._cerr = err.value.decode('utf-8')
  def __enter__(self):
    return self
  def __exit__(self, exc_type, exc_value, traceback):
    if (self._zmConn):
      pfu = _lib.zmDisconnect
      pfun.restype = None
      pfun.argtypes = (ctypes.c_void_p,)
      pfun(self._zmConn)

  #############################################################################
  ### Connection with DB

  def isOK(self) -> bool:
    return self._zmConn is not None
  def setErrorCBack(self, ucb):
    """
    Set ERROR callback
    :param ucb: def func(err : str)
    """
    if (self._zmConn):      
      def c_ecb(err: ctypes.c_char_p, udata: ctypes.c_void_p):
        ucb(err.decode("utf-8"))
      
      errCBackType = ctypes.CFUNCTYPE(None, ctypes.c_char_p, ctypes.c_void_p)    
      self._userErrCBack = errCBackType(c_ecb)

      pfun = _lib.zmSetErrorCBack
      pfun.restype = ctypes.c_bool
      pfun.argtypes = (ctypes.c_void_p, errCBackType, ctypes.c_void_p)
      return pfun(self._zmConn, self._userErrCBack, 0)
  def getLastError(self) -> str:
    """
    Last ERROR strind
    :return: errStr
    """
    if (self._zmConn):
      pfun = _lib.zmGetLastError
      pfun.restype = ctypes.c_bool
      pfun.argtypes = (ctypes.c_void_p, ctypes.c_char_p)
      err = ctypes.create_string_buffer(256)
      pfun(self._zmConn, err)
      return err.value.decode("utf-8") 
    return self._cerr
  def createTables(self) -> bool:
    """
    Create tables, will be created if not exist
    :return: True ok
    """
    if (self._zmConn):
      pfun = _lib.zmCreateTables
      pfun.restype = ctypes.c_bool
      pfun.argtypes = (ctypes.c_void_p,)
      return pfun(self._zmConn)
    return False

  
  #############################################################################
  ### Scheduler
  
  def addScheduler(self, iosch : Scheduler) -> bool:
    """
    Add new scheduler
    :param iosch: new Scheduler config
    :return: True - ok
    """
    if (self._zmConn):
      scng = _SchedrCng_C()
      scng.connectPnt = iosch.connectPnt.encode('utf-8')
      scng.capacityTask = iosch.capacityTask
      scng.name = iosch.name.encode('utf-8')
      scng.description = iosch.description.encode('utf-8')
      
      sid = ctypes.c_uint64(0)
      
      pfun = _lib.zmAddScheduler
      pfun.argtypes = (ctypes.c_void_p, _SchedrCng_C, ctypes.POINTER(ctypes.c_uint64))
      pfun.restype = ctypes.c_bool
      if (pfun(self._zmConn, scng, ctypes.byref(sid))):
        iosch.id = sid.value
        return True
    return False
  def getScheduler(self, iosch : Scheduler) -> bool:
    """
    Get Scheduler config by ID
    :param iosch: Scheduler config
    :return: True - ok
    """
    if (self._zmConn):
      scng = _SchedrCng_C()

      sid = ctypes.c_uint64(iosch.id)
      
      pfun = _lib.zmGetScheduler
      pfun.argtypes = (ctypes.c_void_p, ctypes.c_uint64, ctypes.POINTER(_SchedrCng_C))
      pfun.restype = ctypes.c_bool
      if (pfun(self._zmConn, sid, ctypes.byref(scng))):      
        iosch.connectPnt = scng.connectPnt.decode('utf-8')
        iosch.capacityTask = scng.capacityTask
        iosch.name = scng.name.decode('utf-8')
        if scng.description:
          iosch.description = scng.description.decode('utf-8')
          self._freeResources()
        return True
    return False
  def changeScheduler(self, isch : Scheduler) -> bool:
    """
    Change Scheduler config
    :param isch: new Scheduler config
    :return: True - ok
    """
    if (self._zmConn):
      sid = ctypes.c_uint64(isch.id)
      scng = _SchedrCng_C()
      scng.connectPnt = isch.connectPnt.encode('utf-8')
      scng.capacityTask = isch.capacityTask
      scng.name = isch.name.encode('utf-8')
      scng.description = isch.description.encode('utf-8')
      
      pfun = _lib.zmChangeScheduler
      pfun.argtypes = (ctypes.c_void_p, ctypes.c_uint64, _SchedrCng_C)
      pfun.restype = ctypes.c_bool
      return pfun(self._zmConn, sid, scng)
    return False
  def delScheduler(self, schId) -> bool:
    """
    Delete scheduler
    :param schId: Scheduler id
    :return: True - ok
    """
    if (self._zmConn):
      sid = ctypes.c_uint64(schId)
            
    